/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_CORE_OPERATOR_GENERICOPERATOR_H_
#define AIDGE_CORE_OPERATOR_GENERICOPERATOR_H_

#include <memory>
#include <vector>
#include <string>
#include <cassert>
#include <cstring>

#include "aidge/graph/Node.hpp"
#include "aidge/operator/Operator.hpp"
#include "aidge/utils/DynamicAttributes.hpp"
#include "aidge/utils/Registrar.hpp"
#include "aidge/utils/Types.h"


namespace Aidge {
class GenericOperator_Op
    : public Operator,
      public Registrable<GenericOperator_Op, std::string, std::unique_ptr<OperatorImpl>(std::shared_ptr<GenericOperator_Op>)>,
      public DynamicAttributes {
   private:
    using ComputeDimsFunc = std::function<std::vector<std::vector<size_t>>(const std::vector<std::vector<size_t>>&)>;

    IOIndex_t mNbDataIn;
    IOIndex_t mNbIn;
    IOIndex_t mNbOut;
    std::vector<std::shared_ptr<Tensor>> mInputs;
    std::vector<std::shared_ptr<Tensor>> mOutputs;
    ComputeDimsFunc mComputeOutputDims;

   public:
    GenericOperator_Op(const char *type, IOIndex_t nbDataIn, IOIndex_t nbIn, IOIndex_t nbOut)
        : Operator(type), mNbDataIn(nbDataIn), mNbIn(nbIn), mNbOut(nbOut)
    {
        mInputs = std::vector<std::shared_ptr<Tensor>>(nbIn);
        for (std::size_t i = 0; i < nbIn; ++i) {
            mInputs[i] = std::make_shared<Tensor>();
        }
        mOutputs = std::vector<std::shared_ptr<Tensor>>(nbOut);
        for (std::size_t i = 0; i < nbOut; ++i) {
            mOutputs[i] = std::make_shared<Tensor>();
        }
    }

    /**
     * @brief Copy-constructor. Copy the operator attributes and its output tensor(s), but not its input tensors (the new operator has no input associated).
     * @param op Operator to copy.
     */
    GenericOperator_Op(const GenericOperator_Op& op)
        : Operator(op.type().c_str()), mNbDataIn(op.mNbDataIn), mNbIn(op.mNbIn), mNbOut(op.mNbOut)
    {
        // cpy-ctor
        mInputs = std::vector<std::shared_ptr<Tensor>>(mNbIn);
        for (std::size_t i = 0; i < mNbIn; ++i) {
            mInputs[i] = std::make_shared<Tensor>();
        }
        mOutputs = std::vector<std::shared_ptr<Tensor>>(mNbOut);
        for (std::size_t i = 0; i < mNbOut; ++i) {
            mOutputs[i] = std::make_shared<Tensor>(*op.mOutputs[i]);
        }
    }

    /**
     * @brief Clone the operator using its copy-constructor.
     * @see Operator::GenericOperator_Op
     */
    std::shared_ptr<Operator> clone() const override {
        return std::make_shared<GenericOperator_Op>(*this);
    }

    // Helper functions that can be used with setComputeOutputDims():
    static const ComputeDimsFunc Identity;

    void setComputeOutputDims(ComputeDimsFunc func) {
        mComputeOutputDims = func;
    }

    // Override Virtual Opertor methods
    void associateInput(const IOIndex_t inputIdx, std::shared_ptr<Data> data) override final {
        assert(inputIdx < mNbIn && "operators supports only x inputs");

        if (strcmp(data->type(), Tensor::Type) == 0) {
            // TODO: associate input only if of type Tensor, otherwise do nothing
            mInputs[inputIdx] = std::dynamic_pointer_cast<Tensor>(data);
        }
    }

    void computeOutputDims() override final {
        if (mComputeOutputDims) {
            std::vector<std::vector<size_t>> inputsDims(mNbIn, std::vector<size_t>());
            for (std::size_t i = 0; i < mNbIn; ++i) {
                if (mInputs[i]) {
                    inputsDims[i] = mInputs[i]->dims();
                }
            }

            const auto& outputsDims = mComputeOutputDims(inputsDims);
            assert(outputsDims.size() == mNbOut && "The provided ComputeDimsFunc function returns the wrong number of outputs");
            for (std::size_t i = 0; i < mNbOut; ++i) {
                mOutputs[i]->resize(outputsDims[i]);
            }
        }
        else {
            assert(false && "Cannot compute output dim of a GenericOperator");
        }
    }

    bool outputDimsForwarded() const override final {
        if (mComputeOutputDims) {
            return !(mOutputs[0]->empty());
        }
        else {
            assert(false && "GenericOperator cannot forward dims");
            return false;
        }
    }

    std::shared_ptr<Data> getRawInput(const IOIndex_t inputIdx) const override final {
        assert((inputIdx < mNbIn) && "input index out of range for this instance of GenericOperator");
        printf("Info: using getRawInput() on a GenericOperator.\n");
        return std::static_pointer_cast<Data>(mInputs[inputIdx]);
    }

    inline Tensor& input(const IOIndex_t inputIdx) const override final {
        assert((inputIdx < mNbIn) && "input index out of range for this instance of GenericOperator");
        printf("Info: using input() on a GenericOperator.\n");
        return *mInputs[inputIdx];
    }


    std::shared_ptr<Tensor> getInput(const IOIndex_t inputIdx) const override final {
        assert((inputIdx < mNbIn) && "input index out of range for this instance of GenericOperator");
        printf("Info: using getInput() on a GenericOperator.\n");
        return mInputs[inputIdx];
    }
    inline std::shared_ptr<Tensor> getOutput(const IOIndex_t outputIdx) const override final {
        assert((outputIdx < mNbOut) && "output index out of range for this instance of GenericOperator");
        printf("Info: using getOutput() on a GenericOperator.\n");
        return mOutputs[outputIdx];
    }


    std::shared_ptr<Data> getRawOutput(const IOIndex_t outputIdx) const override final {
        assert((outputIdx < mNbOut) && "output index out of range for this instance of GenericOperator");
        printf("Info: using getRawOutput() on a GenericOperator.\n");
        return std::static_pointer_cast<Data>(mOutputs[outputIdx]);
    }

    Tensor& output(const IOIndex_t outputIdx) const override final {
        assert((outputIdx < mNbOut) && "output index out of range for this instance of GenericOperator");
        printf("Info: using output() on a GenericOperator.\n");
        return *mOutputs[outputIdx];
    }

    ~GenericOperator_Op() = default;

    void setBackend(const std::string & /*name*/) override { printf("setBackend: not available yet.\n"); }
    void setDatatype(const DataType & /*datatype*/) override { printf("setDatatype: not available yet.\n"); }
    void forward() override final {
        if(mImpl){
            mImpl->forward();
        }else{
            printf("forward: No implementation is linked.\n");
        }
    }
    void backward() override final {
        if(mImpl){
            mImpl->backward();
        }else{
            printf("backward: No implementation is linked.\n");
        }
    }
    inline IOIndex_t nbInputs() const noexcept override final { return mNbIn; };
    inline IOIndex_t nbDataInputs() const noexcept override final { return mNbDataIn; };
    inline IOIndex_t nbOutputs() const noexcept override final { return mNbOut; };
};

/**
 * @brief Fictive custom operator not associated with any implementation.
 * Allows to import unknown operators and simulate new ones.
 * @param type Type of the fictive operator.
 * @param nbDataIn Number of input data.
 * @param nbIn Number input data + number of learnt parameters.
 * @param nbOut Number of output data.
 * @param name (optional) name of the Operator.
 * @return std::shared_ptr<Node> Node associated with the Generic Operator.
 */
inline std::shared_ptr<Node> GenericOperator(const char *type, IOIndex_t nbDataIn, IOIndex_t nbIn, IOIndex_t nbOut,
                                             const std::string& name = "") {
    return std::make_shared<Node>(std::make_shared<GenericOperator_Op>(type, nbDataIn, nbIn, nbOut), name);
}
}  // namespace Aidge

#endif /* AIDGE_CORE_OPERATOR_GENERICOPERATOR_H_ */
