/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_CORE_OPERATOR_MAXPOOLING_H_
#define AIDGE_CORE_OPERATOR_MAXPOOLING_H_

#include <array>
#include <numeric>
#include <vector>
#include <cmath>

#include "aidge/data/Tensor.hpp"
#include "aidge/graph/Node.hpp"
#include "aidge/operator/Operator.hpp"
#include "aidge/operator/Producer.hpp"
#include "aidge/utils/StaticAttributes.hpp"
#include "aidge/utils/Registrar.hpp"
#include "aidge/utils/Types.h"

namespace Aidge {
enum class MaxPoolingAttr { StrideDims, KernelDims, CeilMode };

template <DimIdx_t DIM>
class MaxPooling_Op : public Operator,
                public Registrable<MaxPooling_Op<DIM>, std::string, std::unique_ptr<OperatorImpl>(const MaxPooling_Op<DIM> &)>,
                public StaticAttributes<MaxPoolingAttr,
                                       std::array<DimSize_t, DIM>,
                                       std::array<DimSize_t, DIM>,
                                       bool> {
private:
    // FIXME: change accessibility
    std::shared_ptr<Tensor> mInput = std::make_shared<Tensor>();
    const std::shared_ptr<Tensor> mOutput = std::make_shared<Tensor>();

public:
    static constexpr const char *Type = "MaxPooling";

    MaxPooling_Op() = delete;

    using Attributes_ = StaticAttributes<MaxPoolingAttr,
                                             std::array<DimSize_t, DIM>,
                                             std::array<DimSize_t, DIM>,
                                             bool>;
    template <MaxPoolingAttr e>
    using attr = typename Attributes_::template attr<e>;

    constexpr MaxPooling_Op(const std::array<DimSize_t, DIM> &kernel_dims,
                            const std::array<DimSize_t, DIM> &stride_dims = create_array<DimSize_t,DIM>(1),
                            bool ceil_mode = false)
        : Operator(Type),
          Attributes_(attr<MaxPoolingAttr::StrideDims>(stride_dims),
                      attr<MaxPoolingAttr::KernelDims>(kernel_dims),
                      attr<MaxPoolingAttr::CeilMode>(ceil_mode)),
          mOutput(std::make_shared<Tensor>()) {
        setDatatype(DataType::Float32);
    }

    /**
     * @brief Copy-constructor. Copy the operator attributes and its output tensor(s), but not its input tensors (the new operator has no input associated).
     * @param op Operator to copy.
     */
    MaxPooling_Op(const MaxPooling_Op<DIM>& op)
        : Operator(Type),
          Attributes_(op),
          mOutput(std::make_shared<Tensor>(*op.mOutput))
    {
        // cpy-ctor
        setDatatype(op.mOutput->dataType());
        mImpl = op.mImpl ? Registrar<MaxPooling_Op<DIM>>::create(mOutput->getImpl()->backend())(*this) : nullptr;
    }

    /**
     * @brief Clone the operator using its copy-constructor.
     * @see Operator::MaxPooling_Op
     */
    std::shared_ptr<Operator> clone() const override {
        return std::make_shared<MaxPooling_Op<DIM>>(*this);
    }

    void associateInput(const IOIndex_t inputIdx, std::shared_ptr<Data> data) override final {
        assert(inputIdx < 1 && "operators supports only 3 inputs");
        (void) inputIdx; // avoid unused warning
        assert(strcmp(data->type(), Tensor::Type) == 0 && "input data must be of Tensor type");

        mInput = std::dynamic_pointer_cast<Tensor>(data);
    }

    void computeOutputDims() override final {
        if (!mInput->empty()) {
            std::array<DimSize_t, DIM + 2> outputDims = {};

            std::function<float(float)> roundingFunction;
            if (this->template getAttr<MaxPoolingAttr::CeilMode>()) {
                roundingFunction = [](float x) { return std::ceil(x); };
            } else {
                roundingFunction = [](float x) { return std::floor(x); };
            }

            for (std::size_t dim = 0; dim < this->template getAttr<MaxPoolingAttr::KernelDims>().size() ; ++dim) {
                outputDims[dim+2] = 1 + static_cast<DimSize_t>(
                                            roundingFunction(static_cast<float>(mInput->dims()[dim+2] -
                                                                    this->template getAttr<MaxPoolingAttr::KernelDims>()[dim]) /
                                            static_cast<float>(this->template getAttr<MaxPoolingAttr::StrideDims>()[dim])));
            }
            outputDims[1] = mInput->dims()[1];
            outputDims[0] = mInput->dims()[0];
            mOutput->resize(outputDims);
        }
    }

    bool outputDimsForwarded() const override final { return !(mOutput->empty()); }


    inline Tensor& input(const IOIndex_t inputIdx) const override final {
        assert(inputIdx == 0 && "operators supports only 1 inputs");
        (void) inputIdx; // avoid unused warning
        return *(mInput.get());
    }
    inline Tensor& output(const IOIndex_t /*outputIdx*/) const override final { return *(mOutput.get()); }


    inline std::shared_ptr<Tensor> getInput(const IOIndex_t inputIdx) const override final {
        assert(inputIdx == 0 && "MaxPooling Operators supports only 1 inputs");
        (void) inputIdx; // avoid unused warning
        return mInput;
    }
    inline std::shared_ptr<Tensor> getOutput(const IOIndex_t outputIdx) const override final {
        assert(outputIdx == 0 && "MaxPooling Operators has only 1 outputs");
        (void) outputIdx; // avoid unused warning
        return mOutput;
    }


    std::shared_ptr<Data> getRawInput(const IOIndex_t inputIdx) const override final {
        assert(inputIdx == 0 && "operators supports only 1 inputs");
        (void) inputIdx; // avoid unused warning
        return std::static_pointer_cast<Data>(mInput);
    }
    std::shared_ptr<Data> getRawOutput(const IOIndex_t outputIdx) const override final {
        assert(outputIdx == 0 && "operator supports only 1 output");
        (void) outputIdx; // avoid unused warning
        return std::static_pointer_cast<Data>(mOutput);
    }


    void setBackend(const std::string &name) override {
        mImpl = Registrar<MaxPooling_Op<DIM>>::create(name)(*this);
        mOutput->setBackend(name);

        // FIXME: temporary workaround
        mInput->setBackend(name);
    }

    void setDatatype(const DataType &datatype) override {
        mOutput->setDatatype(datatype);

        // FIXME: temporary workaround
        mInput->setDatatype(datatype);
    }

    inline IOIndex_t nbInputs() const noexcept override final { return 1; }
    inline IOIndex_t nbDataInputs() const noexcept override final { return 1; }
    inline IOIndex_t nbOutputs() const noexcept override final { return 1; }
    static const std::vector<std::string> getInputsName(){
        return {"data_input"};
    }
    static const std::vector<std::string> getOutputsName(){
        return {"data_output"};
    }
};

template <std::array<DimSize_t, 1>::size_type DIM>
inline std::shared_ptr<Node> MaxPooling(const std::array<DimSize_t, DIM> &kernel_dims,
                                           const std::string& name = "",
                                           const std::array<DimSize_t, DIM> &stride_dims = create_array<DimSize_t,DIM>(1),
                                           bool ceil_mode=false) {
    static_assert(DIM<=MaxDim,"Too many kernel dimensions required by MaxPooling, not supported");
    return std::make_shared<Node>(std::make_shared<MaxPooling_Op<static_cast<DimIdx_t>(DIM)>>(kernel_dims, stride_dims, ceil_mode), name);
}

// helper with C-style array instead of std::array for kernel_dims to allow automatic template DIM deduction
template <DimSize_t DIM>
inline std::shared_ptr<Node> MaxPooling(
    DimSize_t const (&kernel_dims)[DIM],
    const std::string& name = "",
    const std::array<DimSize_t, DIM> &stride_dims = create_array<DimSize_t,DIM>(1),
    bool ceil_mode = false) {
    static_assert(DIM<=MaxDim,"Too many kernel dimensions required by MaxPooling, not supported");
    return MaxPooling(to_array(kernel_dims), name, stride_dims, ceil_mode);
}
}  // namespace Aidge

namespace {
template <>
const char *const EnumStrings<Aidge::MaxPoolingAttr>::data[] = {"StrideDims", "KernelDims", "CeilMode"};
}

#endif /* AIDGE_CORE_OPERATOR_MAXPOOLING_H_ */
