/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_CORE_OPERATOR_METAOPERATOR_H_
#define AIDGE_CORE_OPERATOR_METAOPERATOR_H_

#include "aidge/operator/Operator.hpp"
#include "aidge/graph/GraphView.hpp"
#include "aidge/graph/OpArgs.hpp"
#include "aidge/scheduler/Scheduler.hpp"

namespace Aidge {
class MetaOperator_Op : public Operator,
                public Registrable<MetaOperator_Op, std::array<std::string, 2>, std::unique_ptr<OperatorImpl>(const MetaOperator_Op &)> {
public:
    std::vector<std::shared_ptr<Tensor>> mInputs;
    std::vector<std::shared_ptr<Tensor>> mOutputs; // These are shared with micro-graph outputs tensors

    // Micro-graph handling:
    std::shared_ptr<GraphView> mGraph; // Meta operator micro-graph
    std::shared_ptr<SequentialScheduler> mScheduler;
    // Need to store an ordored list of input/output operators for the micro-graph,
    // because input/output nodes in a GraphView are unordered.
    // TODO: refactor GraphView to handle ordered input/output?
    std::vector<std::pair<std::shared_ptr<Operator>, IOIndex_t>> mInputOps;
    std::vector<std::pair<std::shared_ptr<Operator>, IOIndex_t>> mOutputOps;

   public:
    MetaOperator_Op(const char *type, const std::shared_ptr<GraphView>& graph,
        std::vector<NodePtr> inputNodes = std::vector<NodePtr>(),
        std::vector<NodePtr> outputNodes = std::vector<NodePtr>());

    /**
     * @brief Copy-constructor. Copy the operator attributes and its output tensor(s), but not its input tensors (the new operator has no input associated).
     * @param op Operator to copy.
     */
    MetaOperator_Op(const MetaOperator_Op& op)
        : Operator(op.type().c_str()),
          mGraph(op.mGraph->clone())
    {
        // cpy-ctor
    }

    /**
     * @brief Clone the operator using its copy-constructor.
     * @see Operator::MatMul_Op
     */
    std::shared_ptr<Operator> clone() const override {
        return std::make_shared<MetaOperator_Op>(*this);
    }

    const std::shared_ptr<GraphView>& getMicroGraph() const {
        return mGraph;
    }

    const std::shared_ptr<SequentialScheduler>& getMicroGraphScheduler() const {
        return mScheduler;
    }

    void associateInput(const IOIndex_t inputIdx, std::shared_ptr<Data> data) override final {
        assert(strcmp(data->type(), Tensor::Type) == 0 && "input data must be of Tensor type");

        const auto& inputOp = mInputOps[inputIdx];
        inputOp.first->associateInput(inputOp.second, data);

        // Associate inputs for custom implementation
        mInputs[inputIdx] = std::dynamic_pointer_cast<Tensor>(data);
    }

    void computeOutputDims() override final {
        // Forward dims of micro-graph
        mGraph->forwardDims();

        // Associate outputs to micro-graph outputs for custom implementation
        for (size_t outputIdx = 0; outputIdx < mOutputOps.size(); ++outputIdx) {
            const auto& outputOp = mOutputOps[outputIdx];
            mOutputs[outputIdx] = outputOp.first->getOutput(outputOp.second);
        }
    }

    bool outputDimsForwarded() const override final { return !(mOutputs[0]->empty()); }


    inline Tensor& input(const IOIndex_t inputIdx) const override final {
        assert(inputIdx < mInputs.size() && "inputIdx out of range");
        return *(mInputs[inputIdx].get());
    }

    inline Tensor& output(const IOIndex_t outputIdx) const override final {
        assert(outputIdx < mOutputs.size() && "outputIdx out of range");
        return *(mOutputs[outputIdx].get());
    }

    inline std::shared_ptr<Tensor> getInput(const IOIndex_t inputIdx) const override final {
        assert(inputIdx < mInputs.size() && "inputIdx out of range");
        return mInputs[inputIdx];
    }

    inline std::shared_ptr<Tensor> getOutput(const IOIndex_t outputIdx) const override final {
        assert(outputIdx < mOutputs.size() && "outputIdx out of range");
        return mOutputs[outputIdx];
    }

    std::shared_ptr<Data> getRawInput(const IOIndex_t inputIdx) const override final {
        assert(inputIdx < mInputs.size() && "inputIdx out of range");
        return std::static_pointer_cast<Data>(mInputs[inputIdx]);
    }

    std::shared_ptr<Data> getRawOutput(const IOIndex_t outputIdx) const override final {
        assert(outputIdx < mOutputs.size() && "outputIdx out of range");
        return std::static_pointer_cast<Data>(mOutputs[outputIdx]);
    }

    void setBackend(const std::string &name) override {
        if (Registrar<MetaOperator_Op>::exists({name, type()})) {
            // A custom implementation exists for this meta operator
            mImpl = Registrar<MetaOperator_Op>::create({name, type()})(*this);
        }

        // The micro-graph should always be set to the right backend, since it
        // shares input/output tensors.
        // Input/output tensors backend are updated here.
        mGraph->setBackend(name);
    }

    void setDatatype(const DataType &datatype) override {
        // The micro-graph should always be set to the right data type, since it
        // shares input/output tensors.
        // Input/output tensors data type are updated here.
        mGraph->setDatatype(datatype);
    }

    inline IOIndex_t nbInputs() const noexcept override final { return mGraph->inputs().size(); }
    inline IOIndex_t nbDataInputs() const noexcept override final { return mGraph->dataInputs().size(); }
    inline IOIndex_t nbOutputs() const noexcept override final { return mGraph->outputs().size(); }

    NbElts_t getNbRequiredData(const IOIndex_t inputIdx) const override;
    NbElts_t getNbConsumedData(IOIndex_t inputIdx) const override;
    NbElts_t getNbProducedData(IOIndex_t outputIdx) const override;

    void updateConsummerProducer() override;
    void forward() override;
    void backward() override {
        assert(false && "not implemented");
    }

};

inline std::shared_ptr<Node> MetaOperator(const char *type,
                                  const std::shared_ptr<GraphView>& graph,
                                  const std::string& name = "",
                                  std::vector<NodePtr> inputNodes = std::vector<NodePtr>(),
                                  std::vector<NodePtr> outputNodes = std::vector<NodePtr>())
{
    return std::make_shared<Node>(std::make_shared<MetaOperator_Op>(type, graph, inputNodes, outputNodes), name);
}
}  // namespace Aidge

#endif /* MetaOperator_H_ */
