/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/


#ifndef AIDGE_ERRORHANDLING_H_
#define AIDGE_ERRORHANDLING_H_

#include <cstdio>
#include <memory>

#define AIDGE_STRINGIZE_DETAIL(x) #x
#define AIDGE_STRINGIZE(x) AIDGE_STRINGIZE_DETAIL(x)

#ifdef NO_EXCEPTION
#define AIDGE_THROW_OR_ABORT(ex, ...) \
do { std::printf(__VA_ARGS__); std::abort(); } while (false)
#else
#include <stdexcept>
#include <memory>
#define AIDGE_THROW_OR_ABORT(ex, ...) \
do { \
    int n = 128; \
    std::unique_ptr<char[]> formatted; \
    formatted.reset(new char[n]); \
    const int len = std::snprintf(formatted.get(), n, __VA_ARGS__); \
    if (len >= n) { \
        formatted.reset(new char[len + 1]); \
        std::snprintf(formatted.get(), len + 1, __VA_ARGS__); \
    }; \
    throw ex(formatted.get()); \
} while (false)
#endif

/**
 * Macro for specified API assertions.
 * Used to check logic directly related to user's inputs.
 * If it asserts, it means an user error.
*/
#define AIDGE_ASSERT(stm, ...) \
if (!(stm)) { printf("Assertion failed: " AIDGE_STRINGIZE(stm) " in " __FILE__ ":%d", __LINE__); \
    AIDGE_THROW_OR_ABORT(std::runtime_error, __VA_ARGS__); }

/**
 * Macro for internal assertions.
 * Used to check internal logic not directly related to API user's inputs.
 * If it asserts, it means a bug.
*/
#define AIDGE_INTERNAL_ASSERT(stm) \
assert((stm) && "Internal assertion failed: " #stm " in " __FILE__ ":" AIDGE_STRINGIZE(__LINE__))

#endif //AIDGE_ERRORHANDLING_H_
