/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#include "aidge/operator/MetaOperator.hpp"
#include "aidge/utils/ErrorHandling.hpp"

Aidge::MetaOperator_Op::MetaOperator_Op(const char *type, const std::shared_ptr<GraphView>& graph,
    std::vector<NodePtr> inputNodes,
    std::vector<NodePtr> outputNodes)
    : Operator(type),
        mGraph(graph)
{
    mInputs = std::vector<std::shared_ptr<Tensor>>(mGraph->inputs().size());
    for (std::size_t i = 0; i < mInputs.size(); ++i) {
        mInputs[i] = std::make_shared<Tensor>();
    }
    mOutputs = std::vector<std::shared_ptr<Tensor>>(mGraph->outputs().size());
    for (std::size_t i = 0; i < mOutputs.size(); ++i) {
        mOutputs[i] = std::make_shared<Tensor>();
    }

    // Fill inputsNodes and outputsNodes when there is no ambiguity
    if (inputNodes.empty()) {
        AIDGE_ASSERT(mGraph->inputNodes().size() == 1, "need to specify internal nodes input mapping");
        inputNodes.push_back(*mGraph->inputNodes().begin());
    }

    if (outputNodes.empty()) {
        AIDGE_ASSERT(mGraph->outputNodes().size() == 1, "need to specify internal nodes output mapping");
        outputNodes.push_back(*mGraph->outputNodes().begin());
    }

    AIDGE_ASSERT(mGraph->inputNodes().size() == inputNodes.size(), "wrong number of specified input nodes");
    AIDGE_ASSERT(mGraph->outputNodes().size() == outputNodes.size(), "wrong number of specified output nodes");

    // Identify inputs that are outside the micro-graph
    for (const auto& inputNode : inputNodes) {
        AIDGE_ASSERT(mGraph->inView(inputNode), "input node must be in the graph");
        const std::vector<std::pair<std::shared_ptr<Node>, IOIndex_t>> inputNodeinputs =
            inputNode->inputs();
        
        int inputIdx = 0;   // input idx relative to the current node
        for (const auto& in : inputNodeinputs) {
            if (in.first == nullptr || !mGraph->inView(in.first)) {
                // The input is not connected inside the micro-graph
                // (no connection to this input or connection outside the micro-graph)
                // => it is therefore an input for the meta-operator
                mInputOps.push_back(std::make_pair(inputNode->getOperator(), inputIdx));
            }

            ++inputIdx;
        }
    }

    // The outputs of the output nodes are also the outputs of the meta-operator
    for (const auto& outputNode : outputNodes) {
        AIDGE_ASSERT(mGraph->inView(outputNode), "output node must be in the graph");
        const std::vector<std::vector<std::pair<std::shared_ptr<Node>, IOIndex_t>>> outputNodeoutputs =
            outputNode->outputs();

        for (size_t outputIdx = 0; outputIdx < outputNodeoutputs.size(); ++outputIdx) {
            mOutputOps.push_back(std::make_pair(outputNode->getOperator(), outputIdx));
        }
    }

    AIDGE_INTERNAL_ASSERT(mInputOps.size() == mGraph->inputs().size());
    AIDGE_INTERNAL_ASSERT(mOutputOps.size() == mGraph->outputs().size());
}

Aidge::NbElts_t Aidge::MetaOperator_Op::getNbRequiredData(const IOIndex_t inputIdx) const {
    if (mImpl) {
        return mImpl->getNbRequiredData(inputIdx);
    }
    else {
        const auto& inputOp = mInputOps[inputIdx];
        return inputOp.first->getNbRequiredData(inputOp.second);
    }
}

Aidge::NbElts_t Aidge::MetaOperator_Op::getNbConsumedData(IOIndex_t inputIdx) const {
    if (mImpl) {
        return mImpl->getNbConsumedData(inputIdx);
    }
    else {
        const auto& inputOp = mInputOps[inputIdx];
        return inputOp.first->getNbConsumedData(inputOp.second);
    }
}

Aidge::NbElts_t Aidge::MetaOperator_Op::getNbProducedData(IOIndex_t outputIdx) const {
    if (mImpl) {
        return mImpl->getNbProducedData(outputIdx);
    }
    else {
        const auto& outputOp = mOutputOps[outputIdx];
        return outputOp.first->getNbProducedData(outputOp.second);
    }
}

void Aidge::MetaOperator_Op::updateConsummerProducer() {
    if (mImpl) {
        mImpl->updateConsummerProducer();
    }
    else {
        if (!mScheduler) {
            // Lazy initialization
            mScheduler = std::make_shared<SequentialScheduler>(mGraph);
        }
        
        // TODO: check that generateScheduling() can be called multiple time to iteratively update the schedule.
        // It could be a good idea to unify updateConsummerProducer() and generateScheduling() into a "updateScheduling()"
        mScheduler->generateScheduling();
    }
}

void Aidge::MetaOperator_Op::forward() {
    if (mImpl) {
        // A custom implementation exists for this meta operator
        mImpl->forward();
    }
    else {
        // No custom implementation, use the individual operators implementations
        if (!mScheduler) {
            // Lazy initialization
            // TODO: should we assert that a scheduler already exists at this point?
            // => should be created in updateConsummerProducer()
            mScheduler = std::make_shared<SequentialScheduler>(mGraph);
            mScheduler->generateScheduling();
        }

        mScheduler->forward(false);
    }
}
